@@ -1,7 +1,8 @@
 import spmatrix
+import numpy
 
 def poisson1d(n):
-    L = spmatrix.ll_mat(n, n)
+    L = spmatrix.ll_mat(n, n, 3*n-2)
     for i in range(n):
         L[i,i] = 2
         if i > 0:
@@ -9,9 +10,9 @@ def poisson1d(n):
         if i < n-1:
             L[i,i+1] = -1
     return L
- 
+
 def poisson1d_sym(n):
-    L = spmatrix.ll_mat_sym(n)
+    L = spmatrix.ll_mat_sym(n, 2*n-1)
     for i in range(n):
         L[i,i] = 2
         if i > 0:
@@ -19,7 +20,8 @@ def poisson1d_sym(n):
     return L
         
 def poisson2d(n):
-    L = spmatrix.ll_mat(n*n, n*n)
+    n2 = n*n
+    L = spmatrix.ll_mat(n2, n2, 5*n2-4*n)
     for i in range(n):
         for j in range(n):
             k = i + n*j
@@ -35,7 +37,8 @@ def poisson2d(n):
     return L
 
 def poisson2d_sym(n):
-    L = spmatrix.ll_mat_sym(n*n)
+    n2 = n*n
+    L = spmatrix.ll_mat_sym(n2, 3*n2-2*n)
     for i in range(n):
         for j in range(n):
             k = i + n*j
@@ -47,11 +50,12 @@ def poisson2d_sym(n):
     return L
 
 def poisson2d_sym_blk(n):
-    L = spmatrix.ll_mat_sym(n*n)
-    I = spmatrix.ll_mat_sym(n)
+    n2 = n*n
+    L = spmatrix.ll_mat_sym(n2, 3*n2-2*n)
+    I = spmatrix.ll_mat_sym(n, n)
     for i in range(n):
         I[i,i] = -1
-    P = spmatrix.ll_mat_sym(n)
+    P = spmatrix.ll_mat_sym(n, 2*n-1)
     for i in range(n):
         P[i,i] = 4
         if i > 0:
