@@ -71,6 +71,10 @@ Availability:
 
 #include "umfpack.h"
 #include "pysparse/spmatrix.h"
+#include "pysparse/ll_mat.h"
+
+#define SYMMETRIC 1    /* Symmetric SpMatrix */
+#define GENERAL   0    /* General   SpMatrix */
 
 /***********************************************************************
  * UMFPackObject definition
@@ -78,12 +82,12 @@ Availability:
 
 typedef struct UMFPackObject {
   PyObject_VAR_HEAD
-  int n;
-  int nnz;
-  double *val;		/* pointer to array of values */
-  int *row;			/* pointer to array of indices */
-  int *ind;			/* pointer to array of indices */
-  void *Numeric;
+  int     n;
+  int     nnz;
+  double *val;                  /* pointer to array of values */
+  int    *row;          /* pointer to array of indices */
+  int    *ind;          /* pointer to array of indices */
+  void   *Numeric;
   double *Control;
   double *Info;
 } UMFPackObject;
@@ -149,7 +153,141 @@ UMFPack_getlists(UMFPackObject *self, PyObject *args) {
     return PyErr_NoMemory();
 }
 
-static char solve_doc[] = "self.solve(b, x, systype)\n\
+static char lunz_doc[] = "self.lunz(): Returns number of nonzeros in factors\n";
+
+static PyObject *UMFPack_Lunz(UMFPackObject *self, PyObject *args) {
+
+  int     nrow, ncol, lnz, unz, nz_udiag, status;
+  void   *Numeric = self->Numeric;
+
+  /* Obtain number of nonzeros in factors */
+  status = umfpack_di_get_lunz(&lnz, &unz, &nrow, &ncol, &nz_udiag, Numeric);
+
+  if( status != UMFPACK_OK ) {
+      switch(status) {
+        case UMFPACK_ERROR_invalid_Numeric_object:
+            PyErr_SetString(PyExc_SystemError,
+                            "Get_Lunz:: Invalid Numeric object");
+            return NULL;
+        case UMFPACK_ERROR_argument_missing:
+            PyErr_SetString(PyExc_SystemError,
+                            "Get_Lunz:: Invalid arguments");
+            return NULL;
+      }
+  }
+
+  return Py_BuildValue("iii", lnz, unz, nz_udiag);
+}
+
+static char lu_doc[] = "self.lu()\n\
+    Returns L and U factors, permutation and scaling information.\n\n\
+    Use: (L, U, P, Q, R, do_recip) = self.lu().\n\
+    The original matrix A is factorized into\n\
+         L U = P R A Q\n\
+    where L is unit lower triangular,\n\
+          U is upper triangular,\n\
+          P and Q are permutation matrices,\n\
+          R is a row-scaling diagonal matrix such that\n\
+           the i-th row of A has been divided    by R[i] if do_recip = True,\n\
+           the i-th row of A has been multiplied by R[i] if do_recip = False.\n\
+    \n\
+    L and U are returned as ll_mat sparse matrices.\n\
+    P, Q and R are returned as NumPy arrays.\
+";
+
+static PyObject *UMFPack_Lu(UMFPackObject *self, PyObject *args) {
+
+  int     nrow, ncol, do_recip, lnz, unz, nz_udiag, status, i, j;
+  int    *Lp, *Lj, *Up, *Ui;
+  double *Lx, *Ux, *D;
+  void   *Numeric = self->Numeric;
+
+  PyArrayObject *P,    *Q,    *R;
+  npy_int       *pval, *qval;
+  npy_double                  *rval;
+  npy_intp       dp[1], dq[1], dr[1];
+
+  PyObject      *Lmat,    *Umat;
+  int            ldims[2], udims[2];
+
+  /* Obtain number of nonzeros in factors */
+  status = umfpack_di_get_lunz(&lnz, &unz, &nrow, &ncol, &nz_udiag, Numeric);
+
+  if( status != UMFPACK_OK ) {
+      switch(status) {
+        case UMFPACK_ERROR_invalid_Numeric_object:
+            PyErr_SetString(PyExc_SystemError,
+                            "Get_Lunz:: Invalid Numeric object");
+            return NULL;
+        case UMFPACK_ERROR_argument_missing:
+            PyErr_SetString(PyExc_SystemError,
+                            "Get_Lunz:: Invalid arguments");
+            return NULL;
+      }
+  }
+
+  /* Allocate memory to hold factors */
+  Lp = (int *)calloc(nrow+1, sizeof(int));
+  Lj = (int *)calloc(lnz, sizeof(int));
+  Lx = (double *)calloc(lnz, sizeof(double));
+
+  Up = (int *)calloc(ncol+1, sizeof(int));
+  Ui = (int *)calloc(unz, sizeof(int));
+  Ux = (double *)calloc(unz, sizeof(double));
+
+  dp[0] = nrow; dq[0] = ncol; dr[0] = nrow;
+  P = (PyArrayObject *)PyArray_SimpleNew(1, dp, NPY_INT);
+  pval = (npy_int *)P->data;
+  
+  Q = (PyArrayObject *)PyArray_SimpleNew(1, dq, NPY_INT);
+  qval = (npy_int *)Q->data;
+
+  D = NULL; /* Force diagonal terms into U */
+
+  R = (PyArrayObject *)PyArray_SimpleNew(1, dr, NPY_DOUBLE);
+  rval = (npy_double *)R->data;
+
+  /* Retrieve factors, permutation vectors and row scaling */
+  status = umfpack_di_get_numeric(Lp, Lj, Lx, Up, Ui, Ux, pval, qval, D,
+                                  &do_recip, rval, Numeric);
+
+  if( status != UMFPACK_OK ) {
+      switch(status) {
+          case UMFPACK_ERROR_out_of_memory:
+            PyErr_SetString(PyExc_SystemError, "Get_Numeric:: out of memory");
+            return NULL;
+          case UMFPACK_ERROR_invalid_Numeric_object:
+            PyErr_SetString(PyExc_SystemError,
+                            "Get_Numeric:: Numeric object is invalid");
+            return NULL;
+      }
+  }
+
+  /* Create LL_Mat structure for factor L and populate it */
+  ldims[0] = nrow; ldims[1] = fmin(nrow,ncol);
+  Lmat = SpMatrix_NewLLMatObject( ldims, GENERAL, lnz );
+
+  for( i=0; i<nrow+1; i++ )
+    for( j=Lp[i]; j<Lp[i+1]; j++ )
+        SpMatrix_LLMatSetItem((LLMatObject *)Lmat, i, Lj[j], Lx[j]);
+
+  free( Lp ); free( Lj ); free( Lx );
+
+  /* Create LL_Mat structure for factor U and populate it */
+  udims[0] = fmin(nrow,ncol); udims[1] = ncol;
+  Umat = SpMatrix_NewLLMatObject( udims, GENERAL, unz );
+
+  for( j=0; j<ncol+1; j++ )
+      for( i=Up[j]; i<Up[j+1]; i++ )
+          SpMatrix_LLMatSetItem((LLMatObject *)Umat, Ui[i], j, Ux[i]);
+
+  free( Up ); free( Ui ); free( Ux );
+  
+  /* Return output data */
+  return Py_BuildValue( "OOOOOi", Lmat, Umat, P, Q, R, do_recip );
+}
+
+static char solve_doc[] = "self.solve(b, x, method, irsteps)\n\
 \n\
 solves linear system of equations.\n\
 \n\
@@ -157,8 +295,8 @@ parameters\n\
 ----------\n\
 \n\
 b        array, right hand side of equation\n\
-x        array, solution vector\n\
-systype  'UMFPACK_A': solve A   * x == b\n\
+x        array, solution vector\n\n\
+method   'UMFPACK_A': solve A   * x == b\n\
          'UMFPACK_At': solve A^T * x == b\n\
          'UMFPACK_Pt_L': solve P^T * L * x == b\n\
          'UMFPACK_L': solve L * x == b\n\
@@ -168,70 +306,76 @@ systype  'UMFPACK_A': solve A   * x == b\n\
          'UMFPACK_U': solve U * x == b\n\
          'UMFPACK_Q_Ut': solve Q * U^T * x == b\n\
          'UMFPACK_Ut': solve U^T * x == b\n\
+\n\
+irsteps  Number of iterative refinement steps to attempt.\n\
 ";
 
 static PyObject *
 UMFPack_solve(UMFPackObject *self, PyObject *args) {
   PyArrayObject *b, *x;
-  char *systype = "UMFPACK_A";
-  int sys, status;
+  char *method = "UMFPACK_A";
+  int res, sys, irsteps = -1, status;
 
-  if (!PyArg_ParseTuple(args, "O!O!|s",
-			&PyArray_Type, &b,
-			&PyArray_Type, &x,
-			&systype))
-    return NULL;
+  res = PyArg_ParseTuple(args, "O!O!|si",
+                               &PyArray_Type, &b,
+                               &PyArray_Type, &x,
+                               &method, &irsteps);
+  if( !res )
+      return NULL;
 
   SPMATRIX_CHECK_ARR_DIM_SIZE(b, 1, self->n);
   SPMATRIX_CHECK_ARR_DIM_SIZE(x, 1, self->n);
 
-  if (strcmp(systype, "UMFPACK_A") == 0)
+  if (strcmp(method, "UMFPACK_A") == 0)
       sys = UMFPACK_A;
-  else if (strcmp(systype, "UMFPACK_At") == 0)
+  else if (strcmp(method, "UMFPACK_At") == 0)
       sys = UMFPACK_At;
-  else if (strcmp(systype, "UMFPACK_Pt_L") == 0)
+  else if (strcmp(method, "UMFPACK_Pt_L") == 0)
       sys = UMFPACK_Pt_L;
-  else if (strcmp(systype, "UMFPACK_L") == 0)
+  else if (strcmp(method, "UMFPACK_L") == 0)
       sys = UMFPACK_L;
-  else if (strcmp(systype, "UMFPACK_Lt_P") == 0)
+  else if (strcmp(method, "UMFPACK_Lt_P") == 0)
       sys = UMFPACK_Lt_P;
-  else if (strcmp(systype, "UMFPACK_Lt") == 0)
+  else if (strcmp(method, "UMFPACK_Lt") == 0)
       sys = UMFPACK_Lt;
-  else if (strcmp(systype, "UMFPACK_U_Qt") == 0)
+  else if (strcmp(method, "UMFPACK_U_Qt") == 0)
       sys = UMFPACK_U_Qt;
-  else if (strcmp(systype, "UMFPACK_U") == 0)
+  else if (strcmp(method, "UMFPACK_U") == 0)
       sys = UMFPACK_U;
-  else if (strcmp(systype, "UMFPACK_Q_Ut") == 0)
+  else if (strcmp(method, "UMFPACK_Q_Ut") == 0)
       sys = UMFPACK_Q_Ut;
-  else if (strcmp(systype, "UMFPACK_Ut") == 0)
+  else if (strcmp(method, "UMFPACK_Ut") == 0)
       sys = UMFPACK_Ut;
   else {
-      PyErr_SetString(PyExc_ValueError, "systype");
+      PyErr_SetString(PyExc_ValueError, "method");
       return NULL;
   }
 
+  if( irsteps != -1 )
+      self->Control[UMFPACK_IRSTEP] = irsteps;
+
   status = umfpack_di_solve(sys, self->ind, self->row, self->val,
-                             (double *)x->data, (double *)b->data, self->Numeric,
-                             self->Control, self->Info);
+                            (double *)x->data, (double *)b->data,
+                            self->Numeric, self->Control, self->Info);
 
-  switch(status) {
-    case UMFPACK_WARNING_singular_matrix:
-      PyErr_SetString(PyExc_SystemError, "singular matrix");
-      return NULL;
-    case UMFPACK_ERROR_out_of_memory:
-      PyErr_SetString(PyExc_SystemError, "insufficient memory");
-      return NULL;
-    case UMFPACK_ERROR_argument_missing:
-      PyErr_SetString(PyExc_SystemError, "one or more argument missing");
-      return NULL;
-    case UMFPACK_ERROR_invalid_system:
-      PyErr_SetString(PyExc_SystemError, "matrix is not square");
-      return NULL;
-    case UMFPACK_ERROR_invalid_Numeric_object:
-      PyErr_SetString(PyExc_SystemError, "invalid Numeric object");
-      return NULL;
+  switch( status ) {
+  case UMFPACK_WARNING_singular_matrix:
+    PyErr_SetString(PyExc_SystemError, "singular matrix");
+    return NULL;
+  case UMFPACK_ERROR_out_of_memory:
+    PyErr_SetString(PyExc_SystemError, "insufficient memory");
+    return NULL;
+  case UMFPACK_ERROR_argument_missing:
+    PyErr_SetString(PyExc_SystemError, "one or more argument missing");
+    return NULL;
+  case UMFPACK_ERROR_invalid_system:
+    PyErr_SetString(PyExc_SystemError, "matrix is not square");
+    return NULL;
+  case UMFPACK_ERROR_invalid_Numeric_object:
+    PyErr_SetString(PyExc_SystemError, "invalid Numeric object");
+    return NULL;
   }
-
+  
   Py_INCREF(Py_None);
   return Py_None;
 }
@@ -241,7 +385,9 @@ UMFPack_solve(UMFPackObject *self, PyObject *args) {
 PyMethodDef UMFPack_methods[] = {
   {"solve", (PyCFunction)UMFPack_solve, METH_VARARGS, solve_doc},
   {"getlists", (PyCFunction)UMFPack_getlists, METH_VARARGS, getlists_doc},
-  {NULL, NULL}			/* sentinel */
+  {"lunz", (PyCFunction)UMFPack_Lunz, METH_VARARGS, lunz_doc},
+  {"lu", (PyCFunction)UMFPack_Lu, METH_VARARGS, lu_doc},
+  {NULL, NULL}          /* sentinel */
 };
 
 
@@ -275,10 +421,10 @@ UMFPack_getattr(UMFPackObject *self, char *name)
     PyObject *list = PyList_New(sizeof(members)/sizeof(char *));
     if (list != NULL) {
       for (i = 0; i < sizeof(members)/sizeof(char *); i ++)
-	PyList_SetItem(list, i, PyString_FromString(members[i]));
+    PyList_SetItem(list, i, PyString_FromString(members[i]));
       if (PyErr_Occurred()) {
-	Py_DECREF(list);
-	list = NULL;
+    Py_DECREF(list);
+    list = NULL;
       }
     }
     return list;
@@ -298,15 +444,15 @@ PyTypeObject UMFPackType = {
   sizeof(UMFPackObject),
   0,
   (destructor)UMFPack_dealloc,   /* tp_dealloc */
-  0,				/* tp_print */
+  0,                /* tp_print */
   (getattrfunc)UMFPack_getattr,  /* tp_getattr */
-  0,				/* tp_setattr */
-  0,				/* tp_compare */
-  0,				/* tp_repr */
-  0,				/* tp_as_number*/
-  0,				/* tp_as_sequence*/
-  0,				/* tp_as_mapping*/
-  0,				/* tp_hash */
+  0,                /* tp_setattr */
+  0,                /* tp_compare */
+  0,                /* tp_repr */
+  0,                /* tp_as_number*/
+  0,                /* tp_as_sequence*/
+  0,                /* tp_as_mapping*/
+  0,                /* tp_hash */
 };
 
 /***********************************************************************
@@ -346,7 +492,7 @@ static void sortcol(int *row, double *val, int n)
 
 static PyObject *
 newUMFPackObject(LLMatObject *matrix, int strategy, double tol2by2, int scale,
-                  double tolpivot, double tolsympivot, int irstep)
+                  double tolpivot, double tolsympivot)
 {
   UMFPackObject *self;
   void *Symbolic;
@@ -410,9 +556,6 @@ newUMFPackObject(LLMatObject *matrix, int strategy, double tol2by2, int scale,
   if (tolsympivot != -1)
     self->Control[UMFPACK_SYM_PIVOT_TOLERANCE] = tolsympivot;
 
-  if (irstep != -1)
-    self->Control[UMFPACK_IRSTEP] = irstep;
-
   self->Info = (double *)PyMem_Malloc(UMFPACK_INFO * sizeof(double));
   if (self->Info == NULL) {
     goto failmemory;
@@ -421,13 +564,13 @@ newUMFPackObject(LLMatObject *matrix, int strategy, double tol2by2, int scale,
   validx = 0;
   for (i=0; i < self->n; i++) {
     self->ind[i] = validx;
-    for (curridx=colidx->root[i]; curridx != -1; curridx = colidx->link[curridx]) {
+    for (curridx=colidx->root[i]; curridx != -1; curridx=colidx->link[curridx]){
       self->val[validx] = matrix->val[curridx];
       self->row[validx++] = colidx->row[curridx];
     }
 
     if (matrix->issym) {
-      for (curridx=matrix->root[i]; curridx != -1; curridx = matrix->link[curridx]) {
+      for (curridx=matrix->root[i]; curridx != -1; curridx=matrix->link[curridx]){
         if (i != matrix->col[curridx]) {
           self->val[validx] = matrix->val[curridx];
           self->row[validx++] = matrix->col[curridx];
@@ -481,7 +624,7 @@ newUMFPackObject(LLMatObject *matrix, int strategy, double tol2by2, int scale,
 
   umfpack_di_free_symbolic(&Symbolic);
 
-  return self;
+  return (PyObject *)self;
 
   failmemory:
     SpMatrix_LLMatDestroyColIndex(&colidx);
@@ -534,8 +677,6 @@ tolpivot            relative pivot tolerance for threshold partial\n\
 tolsympivot         if diagonal pivoting is attempted then this\n\
                     parameter is used to control when the diagonal\n\
                     is selected in a given pivot column.\n\
-\n\
-irstep              number of iterative refinement steps to attempt.\
 ";
 
 static PyObject *
@@ -546,20 +687,18 @@ factorize(PyObject *self, PyObject *args, PyObject *keywds) {
   char *scale="UMFPACK_SCALE_SUM";
   double tolpivot = 0.1;
   double tolsympivot = 0.0;
-  int irstep = 2;
   int res;
-  int strategyval, scaleval;
+  int strategyval = UMFPACK_STRATEGY_AUTO, scaleval = UMFPACK_SCALE_SUM;
 
-  static char *kwlist[] = {"", "strategy", "tol2by2", "scale", "tolpivot", "tolsympivot", "irstep", NULL};
+  static char *kwlist[] = {"", "strategy", "tol2by2", "scale", "tolpivot", "tolsympivot", NULL};
 
-  res = PyArg_ParseTupleAndKeywords(args, keywds, "O!|sdsddi", kwlist, 
-					&LLMatType, &matrix,
-					&strategy,
-					&tol2by2,
-					&scale,
-					&tolpivot,
-					&tolsympivot,
-                              &irstep);
+  res = PyArg_ParseTupleAndKeywords(args, keywds, "O!|sdsdd", kwlist, 
+                    &LLMatType, &matrix,
+                    &strategy,
+                    &tol2by2,
+                    &scale,
+                    &tolpivot,
+                    &tolsympivot);
   if (!res)
     return NULL;
 
@@ -579,7 +718,7 @@ factorize(PyObject *self, PyObject *args, PyObject *keywds) {
   if (strcmp("UMFPACK_SCALE_MAX", scale) == 0)
     scaleval = UMFPACK_SCALE_MAX;
 
-  return newUMFPackObject(matrix, strategyval, tol2by2, scaleval, tolpivot, tolsympivot, irstep);
+  return newUMFPackObject(matrix, strategyval, tol2by2, scaleval, tolpivot, tolsympivot);
 }
 
 
@@ -587,7 +726,7 @@ factorize(PyObject *self, PyObject *args, PyObject *keywds) {
  */
 static PyMethodDef precon_methods[] = {
   {"factorize", (PyCFunction)factorize, METH_VARARGS|METH_KEYWORDS, factorize_doc},
-  {NULL, NULL}	/* sentinel */
+  {NULL, NULL}  /* sentinel */
 };
 
 
